from battlenet_client.cache.cache import Cache
from battlenet_client.wow.base import WoWBase, WoWMedia
from battlenet_client.wow.game_data import pvp_season, pvp_leaderboard, pvp_rewards_index
from battlenet_client.wow.game_data import pvp_tier, pvp_tier_media

import json
from typing import Optional, Union
from requests_oauthlib import OAuth2Session
from oic import oic


class PvPSeason(WoWBase):
    pass


class PvPLeaderboard(WoWBase):
    pass


class PvPRewards(WoWBase):
    pass


class PvPTier(WoWBase):
    pass



"""
"""

# todo: Finish non-account modules first.


from operator import attrgetter
import pendulum
from battlenet_client.wow.profile import profile, specializations_summary, media_summary, encounters
from battlenet_client.wow.profile import achievement_summary, achievement_statistics, pvp
from battlenet_client.wow.realms import Realm
from battlenet_client.wow.playables import PlayableClass, PlayableSpecialization
from battlenet_client.wow.races import PlayableRace


MAX_LEVEL = {'classic1x': 60, 'classic': 80, 'retail': 70}


class Character:
    """Defines a character in World of Warcraft

    Args:
        region_tag (str): region abbreviation for use with the APIs
        realm_tag (str or int): the realm of the character's creation
        name (str): the character name

    Keyword Args:
        release (str, optional): the release to use.  Currently 'classic' only.  The default of None, indicates to use
            the retail data
        locale (str, optional): the localization to use from the API, The default of None means all localizations
        scope (list of str, optional): the scope or scopes to use during the endpoints that require the
            Web Application Flow
        redirect_uri (str, optional): the URI to return after a successful authentication between the user and Blizzard
        client_id (str): the client ID from the developer portal
        client_secret (str): the client secret from the developer portal


    Attributes:
        realm (:obj:`Realm`): the realm instance
        playable_class (:obj:`PlayableClass`): the class of the character
        active_specialization (:obj:`PlayableSpecialization`): the current specialization in use
        race (:obj:`PlayableRace`): the race of the character
        id (int): the ID of the character
        name (str): the properly capitalized name of the character
        gender (str): the gender of the character
        faction (str): the faction the character belongs
        level (int): the level of the character
        experience (int or None): the experience remaining until the next level
        last_login (:obj:`datetime`): the last time the character was active
        achievement_points (int): the total achievement points acquired by the character
            (may include account wide achievement points as well)
        avg_ilevel (int): The average item level (ilevel) with the highest gear either equipped
            or in the character's bags
        equipped_ilevel (int): the item level (ilevel) of the all of the equipped gear
        available_specs (list of :obj:`PlayableSpecialization`): a list of the other
            possible specializations
    """

    def __init__(self, region_tag, realm_tag, name, *, release=None, locale=None, client=None):

        self.playable_class = None
        self.active_specialization = None
        self.race = None
        self.object_id = None
        self.name = name
        self.realm_tag = realm_tag
        self.gender = None
        self.faction = None
        self.level = None
        self._specs = None
        self.equipped_ilevel = None
        self.avg_ilevel = None
        self.achievement_points = None
        self.last_login = None
        self.experience = None
        self.realm = None
        self._media = None
        self.available_specs = None
        self._dungeons = None
        self._raids = None
        self.pvp = None
        self.achievement_summary = None
        self.achievement_statistics = None

        (url, params) = profile(region_tag, realm_tag, name, status=False, release=release, locale=locale)

        if client:
            self.get(client)
        
    def __repr__(self):
        return f"{self.name} ({self.realm.name} {self.realm.region.name})"

    def __str__(self):
        return f"{self.name} ({self.realm.name} {self.realm.region.name})"

    def media(self, media_type='avatar'):
        try:
            return self._media[f'{media_type}_url']
        except KeyError:
            print("media_type needs to be 'avatar', 'bust', 'render', or no value")

    def raid_progression(self, tier=None):
        if not tier:
            try:
                return self._raids['expansions'][-1]
            except KeyError:
                return 'no raid progression'

    def get(self, client):

        api_data = client.get(url, params=params)
        release = "retail"

        if params['namespace'].count("-") > 1:
            start = params['namespace'].find('-')
            end = params['namespace'].find('-', -1)
            release = params['namespace'][start:end]

        self.realm = Realm(region_tag, realm_tag, release=release, locale=params['locale'], client=client)

        self.playable_class = PlayableClass(region_tag, class_id=api_data['character_class']['id'],
                                            release=release, locale=params['locale'], client=client)

        self.active_specialization = PlayableSpecialization(region_tag, spec_id=api_data['active_spec']['id'],
                                                            release=release, locale=params['locale'],
                                                            client=client)
        self.race = PlayableRace(region_tag, race_id=api_data['race']['id'], release=release,
                                 locale=params['locale'], client=client)
        self.object_id = int(api_data['id'])
        self.name = api_data['name']
        self.gender = api_data['gender']['name']
        self.faction = api_data['faction']['name']
        self.level = int(api_data['level'])

        if self.level < MAX_LEVEL[client.release]:
            self.experience = int(api_data['experience'])
        else:
            self.experience = 0

        self.last_login = pendulum.from_timestamp(api_data['last_login_timestamp'] / 1000, tz=self.realm.timezone)
        self.achievement_points = api_data['achievement_points']
        self.avg_ilevel = int(api_data['average_item_level'])
        self.equipped_ilevel = int(api_data['equipped_item_level'])
        self._specs = specializations_summary(client, realm_tag, self.name)
        self._media = media_summary(client, realm_tag, self.name)
        self.available_specs = sorted(
            [PlayableSpecialization(region_tag, spec_id=spec['specialization']['id'], release=release,
                                    locale=params['locale'], client=client)
             for spec in self._specs['specializations']
                if spec['specialization']['id'] != self.active_specialization.id], key=attrgetter('name'))

        # for now, keep these to the API returned endpoints
        self._dungeons = encounters(realm_tag, self.name, 'dungeons')
        self._raids = encounters(realm_tag, self.name, 'raids')
        self.achievement_summary = achievement_summary(client, self.realm.slug, self.name)
        self.achievement_statistics = achievement_statistics(client, self.realm.slug, self.name)
        self.pvp = pvp(client, self.realm.slug, self.name)


"""Auctions

Disclaimer:
All rights reserved, Blizzard is the intellectual property owner of WoW and WoW Classic
and any data pertaining thereto

moduleauthor: David "Gahd" Couples <gahdania@gahd.online>
"""

from battlenet_client import currency_convertor
from base import WoWBase
from battlenet_client.wow.game_data import auction_house



class AuctionHouse(WoWBaseObject):
    """Defines the auction house and organizes the endpoints

    Args:
        region_tag (str): region abbreviation for use with the APIs
        connected_realm_id (int): the ID of the connected realms (a collection of realms acting as one larger realm)

    Keyword Args:
        release (str, optional): the release to use.  Currently 'classic' only.  The default of None, indicates to use
            the retail data
        locale (str, optional): the localization to use from the API, The default of None means all localizations
        scope (list of str, optional): the scope or scopes to use during the endpoints that require the
            Web Application Flow
        redirect_uri (str, optional): the URI to return after a successful authentication between the user and Blizzard
        client_id (str): the client ID from the developer portal
        client_secret (str): the client secret from the developer portal
    """

    def __init__(self, client: region_tag, connected_realm_id, *, release=None, locale=None, client=None):
        super().__init__(region_tag, connected_realm_id, release, locale, client)

        self.connected_realm = None
        self.auctions = None

        if client:
            self.get()

    def __str__(self):
        return f"Auctions: {self.connected_realm.__str__()}"

    def get(self, client=client):
        (url, params) = auction_house(region_tag, self.object_id, release=self.release, locale=self.locale)

        api_data = client.get(url, params=params)

        self.connected_realm = api_data['connected_realm']
        self.auctions = [Auction(auction) for auction in api_data['auction']]


class Auction:
    """Defines the realm specifics

    Args:
        auction (dict): auction data for an item in the auction house

    Attributes:
        id (int): ID of the auction
        item (dict): item endpoints, including any specific endpoints for the item type
        quantity (int): number of the item being sold in this auction
        unit_price (int): cost per item for the auction
        time_left (str): qualitative description of how much time is left
        buyout_price (int): the amount to buyout the auction
    """

    def __init__(self, auction):

        self.object_id = int(auction['id'])
        self.item = auction['item']
        self.quantity = int(auction['quantity'])
        self.unit_price = currency_convertor(auction['unit_price'])
        self.time_left = auction['time_left']

        if 'buyout' in auction.keys():
            self.buyout_price = currency_convertor(auction['buyout'])



"""Defines the Achievement related classes for WoW and WoW Classic

Disclaimer:
All rights reserved, Blizzard is the intellectual property owner of WoW and WoW Classic
and any data pertaining thereto

Author:
David "Gahd" Couples <gahdania@gahd.online>
"""

from typing import Union, Optional
from battlenet_client.wow.game_data import creature_family, creature_type, creature_family_media
from battlenet_client.wow.game_data import creature_display_media, creature


class CreatureFamily:

    def __init__(self, client: Union[OAuth2Session, OicClient], region_tag: str, family_id: Optional[str, int] = None, *,
                 release: Optional[str] = None, locale: Optional[str] = None):

        (url, params) = creature_family(region_tag, family_id, release=release, locale=locale)

        api_data = client.get(url, params=params)

        if not family_id:
            self.creature_families = api_data['creature_families']
        else:
            self.object_id = api_data['id']
            self.name = api_data['name']
            self.spec = api_data['specialization']
            self.media = api_data['media']['key']['href']


class CreatureType:

    def __init__(self, client: Union[OAuth2Session, OicClient], region_tag: str, type_id: Optional[str, int] = None, *,
                 release: Optional[str] = None, locale: Optional[str] = None):

        (url, params) = creature_type(region_tag, type_id, release=release, locale=locale)
        api_data = client.get(url, params=params)

        if not type_id:
            self.creature_types = api_data['creature_types']
        else:
            self.object_id = api_data['id']
            self.name = api_data['name']


class Creature:

    def __init__(self, region_tag, creature_id, *, release=None, locale=None,  client=None):

        self.object_id = None
        self.name = None
        self.type = None
        self.family = None
        self.display_media = None
        self.is_tamable = None

        (url, params) = creature(region_tag, creature_id, release=release, locale=locale)
        api_data = client.get(url, params=params)

        self.object_id = api_data['id']
        self.name = api_data['name']
        self.type = CreatureType(client, api_data['type']['id'], release=release, locale=locale)
        self.family = CreatureFamily(client, api_data['family']['id'], release=release, locale=locale)
        self.display_media = api_data['creature_displays']['key']['href']
        self.is_tamable = bool(api_data['is_tamable'])



"""Defines the classes that handle the game data APIs for Starcraft 2

Functions:
    league_data(region_tag, season_id, queue_id, team_type, league_id, locale)

Misc Variables:
    __version__
    __author__

Author: David "Gahd" Couples
License: GPL v3
Copyright: February 24, 2022
"""
from typing import Optional

from .. import utils
from ..decorators import verify_region


__version__ = '2.0.0'
__author__ = 'David \'Gahd\' Couples'


@verify_region
def league_data(
    region_tag: str,
    season_id: str,
    queue_id: str,
    team_type: str,
    league_id: str,
    locale: Optional[str] = None,
):
    """Returns data for the specified season, queue, team, and league.

    queueId: the standard available queueIds are: 1=WoL 1v1, 2=WoL 2v2, 3=WoL 3v3, 4=WoL 4v4, 101=HotS 1v1,
        102=HotS 2v2, 103=HotS 3v3, 104=HotS 4v4, 201=LotV 1v1, 202=LotV 2v2, 203=LotV 3v3, 204=LotV 4v4,
        206=LotV Archon. Note that other available queues may not be listed here.

    teamType: there are two available teamTypes: 0=arranged, 1=random.

    leagueId: available leagueIds are: 0=Bronze, 1=Silver, 2=Gold, 3=Platinum, 4=Diamond, 5=Master, 6=Grandmaster.

    Args:
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request
        season_id (str): The season ID of the data to retrieve.
        queue_id (str): The queue ID of the data to retrieve.
        team_type (str): The team type of the data to retrieve.
        league_id (str): The league ID of the data to retrieve.

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/sc2/league/{season_id}/{queue_id}/{team_type}/{league_id}"
    params = {"locale": utils.localize(locale)}
    return uri, params



"""World of Warcraft Azerite Essences, Item class, Item subclass, item sets, and related indices

Disclaimer:
All rights reserved, Blizzard is the intellectual property owner of WoW and WoW Classic
and any data pertaining thereto

Author:
David "Gahd" Couples <gahdania@gahd.io>
"""
from battlenet_client.wow.game_data import azerite_essence, item_class, item_set, item_subclass, item
from battlenet_client.wow.playables import PlayableSpecialization


class AzeriteEssence:
    """Index of Azerite Essences

    Args:
        client (object): OIDC/OAuth compliant client
        region_tag (str): region abbreviation for use with the APIs
        release (str, optional): the release to use.
        locale (str, optional): the localization to use from the API, The default of None means all localizations


    Attributes:
        locale (str): the localization
        azerite_essences (list of dict): the index of essences
        name (str): name of a single azerite essence
        essence_id (int): single essence ID
        allowed_specs (list):  the specializations that are able to use the azerite essence
        powers (list):  powers granted by the azerite essence
        media (str): URL to the media for the essence
    """

    def __init__(self, client: Union[OAuth2Session, OicClient], region_tag: str, essence_id: Optional[int] = None, *,
                 release: Optional[str] = None, locale: Optional[str] = None):

        (url, params) = azerite_essence(region_tag, essence_id, release=release, locale=locale)
        api_data = client.get(url, params=params)

        if not essence_id:
            self.azerite_essences = api_data['azerite_essences']
        else:
            self.essence_id = api_data['id']
            self.name = api_data['name']
            self.allowed_specs = [PlayableSpecialization(client, region_tag, spec['id'], release=release, locale=locale)
                                  for spec in api_data['allowed_specializations']]
            self.powers = api_data['powers']
            self.media = api_data['media']['key']['href']


class ItemClass:
    """Index of Item Classes

    Args:
        region_tag (str): region abbreviation for use with the APIs

    Keyword Args:
        release (str, optional): the release to use.
        locale (str, optional): the localization to use from the API, The default of None means all localizations
        scope (list of str, optional): the scope or scopes to use during the endpoints that require the
            Web Application Flow
        redirect_uri (str, optional): the URI to return after a successful authentication between the user and Blizzard
        client_id (str): the client ID from the developer portal
        client_secret (str): the client secret from the developer portal

    Attributes:
        locale (str): the localization
        item_classes (list of dict): the index of item classes
    """

    def __init__(self, client: Union[OAuth2Session, OicClient], region_tag: str, item_class_id: Optional[str, int] = None, *,
                 release: Optional[str] = None, locale: Optional[str] = None):

        (url, params) = item_class(region_tag, item_class_id, release=release, locale=locale)
        client.get(url, params=params)

        if not item_class_id:
            self.item_classes = api_data['item_classes']
        else:
            self.item_class_id = api_data['id']
            self.name = api_data['name']
            self.subclasses = api_data['item_subclasses']


class ItemSet:
    """Index of Item Sets

    Args:
        region_tag (str): region abbreviation for use with the APIs

    Keyword Args:
        release (str, optional): the release to use.
        locale (str, optional): the localization to use from the API, The default of None means all localizations
        scope (list of str, optional): the scope or scopes to use during the endpoints that require the
            Web Application Flow
        redirect_uri (str, optional): the URI to return after a successful authentication between the user and Blizzard
        client_id (str): the client ID from the developer portal
        client_secret (str): the client secret from the developer portal

    Attributes:
        locale (str): the localization
        item_sets (list of dict): the index of item sets
    """

    def __init__(self, client: Union[OAuth2Session, OicClient], region_tag: str, set_id: Optional[str, int] = None, *,
                 release: Optional[str] = None, locale: Optional[str] = None):

        (url, params) = item_set(region_tag, set_id, release=release, locale=locale)
        api_data = client.get(url, params=params)

        if not set_id:
            self.item_sets = api_data['item_sets']
        else:
            self.object_id = api_data['id']
            self.name = api_data['name']
            self.items = api_data['items']


class ItemSubClass:
    """Index of the Item Subclasses.

    Args:
        region_tag (str): region abbreviation for use with the APIs

    Keyword Args:
        release (str, optional): the release to use.
        locale (str, optional): the localization to use from the API, The default of None means all localizations
        scope (list of str, optional): the scope or scopes to use during the endpoints that require the
            Web Application Flow
        redirect_uri (str, optional): the URI to return after a successful authentication between the user and Blizzard
        client_id (str): the client ID from the developer portal
        client_secret (str): the client secret from the developer portal

    Attributes:
        locale (str): the localization
        subclasses (list of dict): the index of races
    """

    def __init__(self, client: Union[OAuth2Session, OicClient], region_tag: str, subclass_id: Optional[str, int] = None, *,
                 release: Optional[str] = None, locale: Optional[str] = None):

        (url, params) = item_subclass(region_tag, subclass_id, release=release, locale=locale)

        api_data = client.get(url, params)

        if not subclass_id:
            self.subclasses = api_data['item_subclasses']
        else:
            self.class_id = api_data['class_id']
            self.subclass_id = api_data['subclass_id']
            self.name = api_data['display_name']
            self.hide = bool(api_data['hide_subclass_in_tooltips'])


class Item:
    """Playable Race

    Args:
        region_tag (str): region abbreviation for use with the APIs
        item_id (int): race ID

    Keyword Args:
        release (str, optional): the release to use.
        locale (str, optional): the localization to use from the API, The default of None means all localizations
        scope (list of str, optional): the scope or scopes to use during the endpoints that require the
            Web Application Flow
        redirect_uri (str, optional): the URI to return after a successful authentication between the user and Blizzard
        client_id (str): the client ID from the developer portal
        client_secret (str): the client secret from the developer portal

    Attributes:
        locale (str): localization
        id (int): the playable race's ID
        name (str): item's name
        quality (str): the quality level
        required_level (int): the minimum level to use the item
        item_class (dict): the item's class
        item_subclass (dict): the item's subclass
        inventory_type (str): item's inventory type
        purchase_price (dict): purchase price in gold, silver, and copper
        sell_price (dict): selling price in gold, silver, and copper
        max_count (int): the max number of the item allowed to owned by a character
        is_equippable (bool): can the item be equipped
        is_stackable (bool): can the item be stacked
        media (str): URL of the media
    """

    def __init__(self, client: Union[OAuth2Session, OicClient], region_tag: str, item_id: Union[str, int] = None, *,
                 release: Optional[str] = None, locale: Optional[str] = None):

        (url, params) = item(region_tag, item_id, release=release, locale=locale)
        
        api_data = client.get(url, params=params)

        self.object_id = api_data['id']
        self.name = api_data['name']
        self.quality = api_data['quality']['name']
        self.level = api_data['level']
        self.required_level = api_data['required_level']
        self.item_class = api_data['item_class']
        self.item_subclass = api_data['item_subclass']
        self.inventory_type = api_data['inventory_type']['name']
        self.purchase_price = self._client.currency_convertor(api_data['purchase_price'])
        self.sell_price = self._client.currency_convertor(api_data['sell_price'])
        self.max_count = api_data['max_count']
        self.is_equippable = bool(api_data['is_equippable'])
        self.is_stackable = bool(api_data['is_stackable'])
        self.media = api_data['media']['key']['href']



"""World of Warcraft Connected Realms and Realm


Classes:
    ConnectedRealm
    Realm
"""

from battlenet_client.wow.base import WoWBase
from battlenet_client.cache.cache import Cache
from battlenet_client.wow.regions import Region
from battlenet_client.wow.game_data import connected_realm, realm
from battlenet_client.utils import slugify

import json
from oic import oic
from requests_oauthlib import OAuth2Session
from typing import Optional, Union, List, Dict


class ConnectedRealm(WoWBase):
    """World of Warcraft Connected Realm

    Connected Realms are multiple instances of :obj:`Realm` where the players can interact with one another,
    as if they are on the same realm.

    Attributes:
        region (:obj:`Region`): The region of the connected realm
        status (str): status of the connected realms
        has_queue (bool):  indication that the connected realms have a queue or not
        population (str): population of the server
        realms (list of :obj:`Realm`): list of :obj:`Realms` that make up this connected realm
        connected_realm_id (int): ID for the connected realm
    """

    def __init__(self, client: Union[OAuth2Session, oic.Client], region_tag: str, conn_realm_id: int, *,
                 release: Optional[str] = 'retail', locale: Optional[str] = None, cache: Optional[Cache] = None):
        """Initializes an instance of :obj:`ConnectedRealm`

        Args:
            client (:obj:`OAuth2Session`, :obj:`oic.Client`): used with communications to the API
            region_tag (str): region client is requesting data for
            connected_realm_id (int): requested connected realm ID

        Keyword Args:
            release (str, optional): the release of the game to lookup, defaults to `retail`
            locale (str, optional): the localization to use for the responses, defaults to all
            cache (:obj:`Cache`, optional):

        Raises
            BNetValueError: when the API returns HTTP 404
        """
        (url, params) = connected_realm(region_tag, conn_realm_id, release=release)
        api_data = self.get(client, release, 'connected_realm', url, params, object_id=conn_realm_id,
                            cache=cache)

        super().__init__(region_tag,  api_data['id'], release=release, locale=locale)
        self._status = api_data['status']['name']
        self.has_queue = api_data['has_queue']
        self._population = api_data['population']['name']
        self._realms = [Realm(client, region_tag, obj['id'], release=release, locale=locale, cache=cache)
                        for obj in api_data['realms']]

    def __str__(self):
        return f"{self.__class__.__name__} ({self.realms[0][1][self.locale]} {self.region_tag.upper()})"
    
    def __repr__(self):
        return f"{self.__class__.__name__} ({self.realms[0][1][self.locale]} {self.region_tag.upper()})"

    def __len__(self):
        return len(self.realms)

    @property
    def get_region(self) -> Region:
        """Retrieves the :obj:`Region` of the :obj:`ConnectedRealm`

        Returns:
            (:obj:`Region`): returns the :obj:`Region` for this :obj:`ConnectedRealm`
        """
        return self.realms[0].region

    def get_status(self, locale: Optional[str] = None) -> str:
        """Retrieves the localized status of the :obj:`ConnectedRealm`

        Args:
            locale (str, optional): the localization to retrieve. Defaults to locale of instance

        Returns:
            (str) Localized name of the :obj:`ConnectedRealm`
        """
        return self._localized_data('_status', locale)

    def get_population(self, locale: Optional[str] = None) -> str:
        """Retrieves the localized population of the :obj:`ConnectedRealm`

        Args:
            locale (str, optional): the localization to retrieve. Defaults to locale of instance

        Returns:
            (str) Localized population of the :obj:`ConnectedRealm`
        """
        return self._localized_data('_population', locale)

    def get_realm(self, realm_id: int) -> Realm:
        """Returns if the specified realm IDs is part of this :obj:`ConnectedRealm`

        Args:
            realm_id (int): ID of the realm

        Returns:
            :obj:`Realm`: realm identified by its ID or None if not found
        """
        return self.get_object('_realms', realm_id)


class Realm(WoWBase):
    """Realm of WoW

    Attributes:
        realm_id: (int):  realm's actual id number
        slug (str): simplified string representing the realm
        names (dict of str: str): all the localized names of the realm
        timezone (str): timezone for the realm
        realm_types (str): realm type
        realm_locale (str): realm preferred locale
        category (str): realm's physical country
        is_tournament (bool):  realm used in tournament play
        region (:obj:`Region`): region the realm occupies
    """

    def __init__(self, client: Union[OAuth2Session, oic.Client], region_tag: str, realm_id: Union[str, int], *,
                 release: Optional[str] = 'retail', locale: Optional[str] = None, cache: Optional[Cache] = None):
        """Initializes an instance of :obj:`Realm`

        Args:
            client (:obj:`OAuth2Session`, :obj:`oic.Client`): used with communications to the API
            region_tag (str): region client is requesting data for
            realm_id (int, str): realm's ID or slug (simplified name)

        Keyword Args:
            release (str, optional): the release of the game to lookup, defaults to `retail`
            locale (str, optional): the localization to use for the responses, defaults to all
            cache (:obj:`Cache`, optional):

        Raises
            BNetValueError: when the API returns HTTP 404
        """

        if isinstance(realm_id, str):
            realm_id = slugify(realm_id)

        (url, params) = realm(region_tag, realm_id, release=release)
        api_data = self.get(client, self.release, 'realm', url, params, object_id=realm_id, cache=cache)

        super().__init__(region_tag, api_data['id'], release=release, locale=locale)
        self._name = api_data['name']
        self.slug = api_data['slug']
        self.timezone = api_data['timezone']
        self._realm_type = api_data['type']['name']
        self.realm_locale = api_data['locale']
        self._category = api_data['category']
        self.is_tournament = api_data['is_tournament']
        self.region_id = Region(client, region_tag, api_data['region']['id'], release=release, locale=locale,
                                cache=cache)

    def __str__(self):
        return f"{self.name(self.locale) if self.locale else self.name('enus')} ({self.region_tag.upper()})"

    def __repr__(self):
        return f"{self.name(self.locale) if self.locale else self.name('enus')} ({self.region_tag.upper()})"

    def get_name(self, locale: Optional[str] = None) -> str:
        """Retrieves the localized name of the :obj:`Realm` instance

        Args:
            locale (str, optional): the localization to retrieve. Defaults to locale of instance

        Returns:
            (str) Localized name of the :obj:`Realm`
        """
        return self._localized_data('_name', locale)

    def get_realm_type(self, locale: Optional[str] = None) -> str:
        """Retrieves the localized type of the :obj:`Realm` instance

        Args:
            locale (str, optional): the localization to retrieve. Defaults to locale of instance

        Returns:
            (str) Localized type of the :obj:`Realm`
        """
        return self._localized_data('_realm_type', locale)

    def get_category(self, locale: Optional[str] = None) -> str:
        """Retrieves the localized category of the :obj:`Realm` instance

        Args:
            locale (str, optional): the localization to retrieve. Defaults to locale of instance

        Returns:
            (str) Localized category of the :obj:`Realm`
        """
        return self._localized_data('_category', locale)



"""Defines the base class "BNetClient"

Classes:
    BNetClient
    CredentialClient


Author: David "Gahd" Couples
License: GPL v3
Copyright: February 24, 2022
"""
from typing import Optional, List

from oic import rndstr

from oic.oic import Client
from oic.utils.authn.client import CLIENT_AUTHN_METHOD
from oic.oic.message import RegistrationResponse, AuthorizationResponse
from oic.oauth2 import REQUEST2ENDPOINT
from oic.oauth2.message import CCAccessTokenRequest, ASConfigurationResponse

from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session

from decouple import config

from battlenet_client import utils
from battlenet_client import constants as bnet_constants
from battlenet_client import exceptions as bnet_exceptions


class BNetClient(Client):
    """Defines the base OAuth v2 Authorization code flow/OpenID Connect
    client class for the battlenet_client package

    Args:
        region (str): region abbreviation for use with the APIs

    Keyword Args:
        client_id (str): the client ID from the developer portal
        client_secret (str): the client secret from the developer portal
        scope (list, optional): the scope or scopes to use during the
            endpoints that require the Web Application Flow
        redirect_uri (str, optional): the URI to return after a successful
            authentication between the user and Blizzard

    Attributes:
        tag (str): the region tag (abbreviation) of the client
    """

    __author__ = 'David "Gahd" Couples'
    __version__ = "3.0.0"

    def __init__(self, region: str, *, client_id: Optional[str] = None, client_secret: Optional[str] = None,
                 scope: Optional[List[str]] = None, redirect_uri: Optional[str] = None) -> None:

        if not client_id:
            client_id = config("CLIENT_ID")

        super().__init__(client_id=client_id, client_authn_method=CLIENT_AUTHN_METHOD)

        if not client_secret:

            client_secret = config("CLIENT_SECRET")

        self.client_secret = client_secret

        self.nonce = rndstr()
        self.state = rndstr()

        try:
            self.tag = getattr(bnet_constants, region)
        except AttributeError:
            if region.strip().lower() in ("us", "eu", "kr", "tw", "cn"):
                self.tag = region.strip().lower()
            else:
                raise bnet_exceptions.BNetRegionNotFoundError("Region not available")

        op_data = {"version": "3.0", "issuer": f"{utils.auth_host(region)}/oauth",
                   "token_endpoint": f"{utils.auth_host(region)}/oauth/token", "redirect_uri": redirect_uri}

        if scope:
            op_data.update({"authorization_endpoint": f"{utils.auth_host(region)}/oauth/authorize",
                            "userinfo_endpoint": f"{utils.auth_host(region)}/oauth/userinfo",
                            "jwks_uri": f"{utils.auth_host(region)}/oauth/jwks/certs",
                            'scope': scope})

        else:
            REQUEST2ENDPOINT["CCAccessTokenRequest"] = f"{utils.auth_host(region)}/oauth/token"

        op_info = ASConfigurationResponse(**op_data)
        self.handle_provider_config(op_info, op_info["issuer"])

        info = {
            "client_id": client_id,
            "client_secret": client_secret,
        }
        client_reg = RegistrationResponse(**info)
        self.store_registration_info(client_reg)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} {self.__version__}"

    def __str__(self) -> str:
        return f"{self.__class__.__name__} {self.__version__}"

    def authorization_url(self, **kwargs):
        """Prepares and returns the authorization URL to the Battle.net
        authorization servers

        Returns:
            str: the URL to the Battle.net authorization server
        """
        args = {
            "client_id": self.client_id,
            "response_type": "code",
            "scope": self.provider_info['scope'],
            "nonce": self.nonce,
            "redirect_uri": self.provider_info["redirect_uri"],
            "state": self.state
        }

        auth_req = self.construct_AuthorizationRequest(request_args=args, **kwargs)
        return auth_req.request(self.authorization_endpoint)

    def authorization_response(self, response: str):
        """Processes the authorization response from the Battle.net
        authorization servers

        Args:
            response (str): the query string from the callback

        Returns:
            Nothing
        """
        auth_resp = self.parse_response(AuthorizationResponse, info=response, sformat="urlencoded")
        assert auth_resp["state"] == self.state

        args = {"code": auth_resp["code"]}

        self.do_access_token_request(
            state=auth_resp["state"], request_args=args, authn_method="client_secret_basic"
        )


class CredentialClient(OAuth2Session):
    """Defines the child OAuth v2 Authorization code flow/OpenID Connect client for handling
    Hearthstone API endpoints

    Args:
        region (str): region abbreviation for use with the APIs

    Keyword Args:
        client_id (str): the client ID from the developer portal
        client_secret (str): the client secret from the developer portal

    Attributes:
        tag (str): the region tag (abbreviation) of the client
        api_host (str): the host to use for accessing the API endpoints
        auth_host (str): the host to use for authentication
        render_host (str): the hose to use for images
    """

    __author__ = 'David "Gahd" Couples'
    __version__ = "2.1.1"

    def __init__(self, region: str, *, client_id: Optional[str] = None, client_secret: Optional[str] = None) -> None:

        self._state = None

        if not client_id:
            client_id = config("CLIENT_ID")

        if not client_secret:
            client_secret = config("CLIENT_SECRET")

        try:
            self.tag = getattr(bnet_constants, region)
        except AttributeError:
            if region.strip().lower() in ("us", "eu", "kr", "tw", "cn"):
                self.tag = region.strip().lower()
            else:
                raise bnet_exceptions.BNetRegionNotFoundError("Region not available")

        self._client_secret = client_secret

        if self.tag == "cn":
            self.api_host = "https://gateway.battlenet.com.cn"
            self.auth_host = "https://oauth.battlenet.com.cn"
            self.render_host = "https://render.worldofwarcraft.com.cn"
        elif self.tag in ("kr", "tw"):
            self.api_host = f"https://{self.tag}.api.blizzard.com"
            self.auth_host = "https://oauth.battle.net"
            self.render_host = f"https://render-{self.tag}.worldofwarcraft.com"
        else:
            self.api_host = f"https://{self.tag}.api.blizzard.com"
            self.auth_host = f"https://oauth.battle.net"
            self.render_host = f"https://render-{self.tag}.worldofwarcraft.com"

        super().__init__(client=BackendApplicationClient(client_id=client_id))
        self.fetch_token(token_url=f"{self.auth_host}/oauth/token", client_id=client_id, client_secret=client_secret)

    def __str__(self) -> str:
        return f"{self.__class__.__name__} {self.__version__} ({self.tag.upper()})"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} {self.__version__} ({self.tag.upper()})"





from battlenet_client.cache.cache import Cache
from battlenet_client.wow.base import WoWBase, WoWMedia
from battlenet_client.wow.game_data import talent_tree, talent, pvp_talent, tech_talent_tree, tech_talent
from battlenet_client.wow.game_data import tech_talent_media

import json
from typing import Optional, Union
from requests_oauthlib import OAuth2Session
from oic import oic


class TalentTree(WoWBase):
    # TODO: to be added in v4.2.0
    pass


class TalentTreeNode(WoWBase):
    # TODO: to be added in v4.2.0
    pass


class Talent(WoWBase):
    """Talents

    Attributes:
        spell (int):  spell ID related to this talent
        playable_class (int): playable class ID that can use this talent
    """
    def __init__(self, client: Union[OAuth2Session, oic.Client], region_tag: str, talent_id: int, *,
                 release: Optional[str] = 'retail', locale: Optional[str] = 'enus', cache: Optional[Cache] = None):

        (url, params) = talent(region_tag, talent_id, release=release)

        api_data = self.get(client, release, 'talent', url, params, object_id=talent_id, cache=cache)

        super().__init__(region_tag, api_data['id'], release=release, locale=locale)

        self.spell = api_data['spell']['id']
        self._rank_descriptions = {rank['id']: rank['description'] for rank in api_data['rank_descriptions']}
        self.playable_class = api_data['playable_class']['id']

    def get_rank_description(self, rank: int, locale: Optional[str] = None) -> str:
        """returns the rank description

        Parameters:
            rank (int): rank ID related to this talent
            locale (str, optional): the locale

        Returns:
            str: the rank int given locale
        """
        locale = locale or self.locale or 'en_US'
        return self.rank_descriptions[rank][locale]


class PvpTalent(WoWBase):
    """PvP Talent

    Attributes:
        spell (int): the spell ID associated with this talent
        playable_spec (int): the playable specialization that can use this talent
        level_unlock (int): the player's level that unlocks this talent
        slots (list of int): the slots that this talent can occupy
    """
    def __init__(self, client: Union[OAuth2Session, oic.Client], region_tag: str, talent_id: int, *,
                 release: Optional[str] = 'retail', locale: Optional[str] = 'enus', cache: Optional[Cache] = None):

        (url, params) = pvp_talent(region_tag, talent_id, release=release)

        api_data = self.get(client, release, 'pvp_talent', url, params, object_id=talent_id, cache=cache)

        super().__init__(region_tag, api_data['id'], release=release, locale=locale)

        self.spell = api_data['spell']['id']
        self.playable_spec = api_data['playable_spec']['id']

        self._descriptions = api_data['description']

        self.level_unlock = api_data['unlock_player_level']
        self.slots = api_data['compatible_slots']

    def get_description(self, locale: Optional[str] = None) -> str:
        """Returns the rank description

        Parameters:
            locale (str, optional): the locale

        Returns:
            str: the rank int given locale
        """
        return self._localized_data('_descriptions', locale)


class TechTalentTree(WoWBase):
    # TODO: to be added in v4.2.0
    pass


class TechTalent(WoWBase):
    """Tech Talents

    Tech talents cover talents that are embedded in equipment during the Battle For Azeroth and Shadowlands expansions

    Attributes:
        talent_tree (int): the talent that the talent is part
        tier (int): the tier of the talent
        display_order (int): the order to display the talent
        prerequisite (int): ID of the prerequisite talent
        icon (:obj:`WoWMedia`): the icon of the talent
    """
    def __init__(self, client: Union[OAuth2Session, oic.Client], region_tag: str, talent_id: int, *,
                 release: Optional[str] = 'retail', locale: Optional[str] = 'enus', cache: Optional[Cache] = None):

        (url, params) = tech_talent(region_tag, talent_id, release=release)
        api_data = self.get(client, release, 'tech-talent', url, params, object_id=talent_id, cache=cache)

        super().__init__(region_tag, api_data['id'], release=release, locale=locale)

        self.talent_tree = api_data['talent_tree']['id']
        self.tier = api_data['tier']
        self.display_order = api_data['display_order']
        self.prerequisite = api_data['prerequisite_talent']['id']
        self._names = api_data['name']
        self._descriptions = api_data['description']
        self._spell_tooltips = api_data['spell_tooltip']

        (url, params) = tech_talent(region_tag, talent_id, release=release)
        api_data = self.get(client, release, 'tech-talent-media', url, params, object_id=talent_id, cache=cache)
        self.icon = WoWMedia(api_data)

    def get_name(self, locale: Optional[str] = None) -> str:
        return self._localized_data('_names', locale)

    def get_description(self, locale: Optional[str] = None) -> str:
        return self._localized_data('_descriptions', locale)

    def get_spell_tooltip(self, locale: Optional[str] = None) -> Dict[str, str]:
        locale = locale or self.locale or 'en_US'
        return {'name': self._spell_tooltips['spell']['name'][locale],
                'description': self._spell_tooltips['spell']['description'][locale],
                'cast_name': self._spell_tooltips['spell']['cast_time'][locale]}




import functools
from ..exceptions import BNetRegionError


def sc2_region(func):

    @functools.wraps(func)
    def wrapper_sc2_region(*args, **kwargs):
        if args[0].lower() == 'cn':
            raise BNetRegionError("Not allowed in region")
        return func(*args, **kwargs)
    return wrapper_sc2_region


def sc2_cn_only(func):
    @functools.wraps(func)
    def wrapper_sc2_region(*args, **kwargs):
        if args[0].lower() != 'cn':
            raise BNetRegionError("Not allowed in region")
        return func(*args, **kwargs)
    return wrapper_sc2_region



"""This module contains the classes for accessing profile related APIs

Functions:
    account_collections(region_tag, category, release, locale)
    account_profile_summary(region_tag, release, locale)
    achievement_statistics(region_tag, realm_name, character_name, release, locale)
    achievement_summary(region_tag, realm_name, character_name, release, locale)
    appearance_summary(region_tag, realm_name, character_name, release, locale)
    collections(region_tag, realm_name, character_name, category, release, locale)
    encounters(region_tag, realm_name, character_name, category, release, locale)
    equipment_summary(region_tag, realm_name, character_name, release, locale)
    guild(region_tag, realm_name, guild_name, category, release, locale)
    hunter_pets_summary(region_tag, realm_name, character_name, release, locale)
    media_summary(region_tag, realm_name, character_name, release, locale)
    mythic_keystone(region_tag, realm_name, character_name, season_id, release, locale)
    professions_summary(region_tag, realm_name, character_name, release, locale)
    profile(region_tag, realm_name, character_name, status, release, locale)
    protected_character_profile_summary(region_tag, realm_id, character_id, release, locale)
    pvp(region_tag, realm_name, character_name, pvp_bracket, release, locale)
    quests(region_tag, realm_name, character_name, completed, release, locale)
    reputations_summary(region_tag, realm_name, character_name, release, locale)
    soulbinds(region_tag, realm_name, character_name, release, locale)
    specializations_summary(region_tag, realm_name, character_name, release, locale)
    statistics_summary(region_tag, realm_name, character_name, release, locale)
    title_summary(region_tag, realm_name, character_name, release, locale)

Misc Variables:
    __version__
    __author__

Author: David "Gahd" Couples
License: GPL v3
Copyright: February 24, 2022
"""
from typing import Optional
from urllib.parse import quote as urlquote

from .utils import namespace
from .. import utils
from ..decorators import verify_region
from ..exceptions import BNetValueError


@verify_region
def account_profile_summary(
    region_tag: str,
    *,
    release: Optional[str] = "retail",
    locale: Optional[str] = None
):
    """Returns a profile summary for an account.

    Because this endpoint provides data about the current logged-in user's World of Warcraft account,
    it requires an access token with the 'wow.profile' scope acquired via the Authorization Code Flow.
    https://develop.battle.net/documentation/guides/using-oauth/authorization-code-flow

    Args:
        region_tag (str): region_tag abbreviation
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/profile/user/wow"
    params = {
        "locale": utils.localize(locale),
        "namespace": namespace(region_tag,  "profile", release)
    }

    return uri, params


@verify_region
def protected_character_profile_summary(
    region_tag: str,
    realm_id: int,
    character_id: int,
    *,
    release: Optional[str] = "retail",
    locale: Optional[str] = None
):
    """Returns a protected profile summary for a character.

    Because this endpoint provides data about the current logged-in user's World of Warcraft account,
    it requires an access token with the 'wow.profile' scope acquired via the Authorization Code Flow.
    https://develop.battle.net/documentation/guides/using-oauth/authorization-code-flow

    Args:
        region_tag (str): region_tag abbreviation
        realm_id (int): the ID for the character's realm
        character_id (int): the ID of character
        release (str): release of the game (ie classi`c1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
            profile summary
    """
    uri = f"{utils.api_host(region_tag)}/profile/user/wow/protected-character/{realm_id}-{character_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": namespace(region_tag,  "profile", release)
    }
    return uri, params


@verify_region
def account_collections(
    region_tag: str,
    *,
    category: Optional[str] = None,
    release: Optional[str] = "retail",
    locale: Optional[str] = None,
):
    """Returns an index of collection types for an account.

    Because this endpoint provides data about the current logged-in user's World of Warcraft account,
    it requires an access token with the 'wow.profile' scope acquired via the Authorization Code Flow.
    https://develop.battle.net/documentation/guides/using-oauth/authorization-code-flow

    Args:
        region_tag (str): region_tag abbreviation
        category (str): 'pets' to retrieve the pet collections, and
            'mounts' to retrieve the mount collection of the account or
            None for both pets and mounts
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    if category and category.lower() not in ('pets', 'mounts', 'toys', 'heirlooms'):
        raise BNetValueError('Invalid Category:  \'pets\' or \'mounts\'')

    uri = f"{utils.api_host(region_tag)}/profile/user/wow/collections"
    if category is not None:
        uri += f"/{category}"
    params = {

        "locale": utils.localize(locale),
        "namespace": namespace(region_tag,  "profile", release)
    }
    return uri, params


@verify_region
def achievement_summary(
    region_tag: str,
    realm_name: str,
    character_name: str,
    *,
    statistics: Optional[bool] = False,
    release: Optional[str] = "retail",
    locale: Optional[str] = None
):
    """Returns a summary of the achievements a character has completed.

    Args:
        region_tag (str): region_tag abbreviation
        realm_name (str): the slug for the character's realm
        character_name (str): name of character
        statistics (bool): Boolean to determine to display the statistics or not
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
    uri += f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/achievements"

    if statistics:
        uri += '/statistics'

    params = {
        "locale": utils.localize(locale),
        "namespace": namespace(region_tag,  "profile", release)
    }
    return uri, params


@verify_region
def appearance_summary(
    region_tag: str,
    realm_name: str,
    character_name: str,
    *,
    release: Optional[str] = "retail",
    locale: Optional[str] = None
):
    """Returns a summary of a character's appearance settings.

    Args:
        region_tag (str): region_tag abbreviation
        realm_name (str): the slug for the character's realm
        character_name (str): name of character
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
    uri += f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/appearance"
    params = {
        "locale": utils.localize(locale),
        "namespace": namespace(region_tag,  "profile", release)
    }
    return uri, params


@verify_region
def collections(
    region_tag: str,
    realm_name: str,
    character_name: str,
    *,
    category: Optional[str] = None,
    release: Optional[str] = "retail",
    locale: Optional[str] = None,
):
    """Returns an index of collection types for a character, a summary of the mounts
    a character has obtained, or a summary of the battle pets a character has obtained.

    Args:
        region_tag (str): region_tag abbreviation
        realm_name (str): the slug for the character's realm
        character_name (str): name of character
        category (str): category to retrieve. options are pets or mounts, or None (default).  None will
            provide both
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    if category and category not in ('pets', 'mounts'):
        raise BNetValueError("Category invalid: options are 'pets', 'mounts'")

    uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
    uri += (
        f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/collections"
    )

    if category:
        if category.lower() not in ("pets", "mounts"):
            raise ValueError("Category needs to pets or mounts")
        uri += f"/{utils.slugify(category)}"

    params = {
        "locale": utils.localize(locale),
        "namespace": namespace(region_tag,  "profile", release)
    }
    return uri, params


@verify_region
def encounters(
    region_tag: str,
    realm_name: str,
    character_name: str,
    *,
    category: Optional[str] = None,
    release: Optional[str] = "retail",
    locale: Optional[str] = None,
):
    """Returns a summary of all of a character's encounters, just dungeon encounters,
    or just raid encounters

    Args:
        region_tag (str): region_tag abbreviation
        realm_name (str): the slug for the character's realm
        character_name (str): name of character
        category (str): category to retrieve.  options are 'dungeons',
            'raids', or None (default).  None will access both dungeons and raids
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
    uri += f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/encounters"

    if category:
        if category.lower() not in ("dungeons", "raids"):
            raise ValueError("Available Categories: None, dungeons and raids")
        uri += f"/{utils.slugify(category)}"

    params = {
        "locale": utils.localize(locale),
        "namespace": namespace(region_tag,  "profile", release)
    }
    return uri, params


@verify_region
def equipment_summary(
    region_tag: str,
    realm_name: str,
    character_name: str,
    *,
    release: Optional[str] = "retail",
    locale: Optional[str] = None
):
    """Returns a summary of the items equipped by a character.

    Args:
        region_tag (str): region_tag abbreviation
        realm_name (str): the slug for the character's realm
        character_name (str): name of character
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
    uri += f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/equipment"

    params = {
        "locale": utils.localize(locale),
        "namespace": namespace(region_tag,  "profile", release)
    }
    return uri, params


@verify_region
def hunter_pets_summary(
    region_tag: str,
    realm_name: str,
    character_name: str,
    *,
    release: Optional[str] = "retail",
    locale: Optional[str] = None
):
    """If the character is a hunter, returns a summary of the character's hunter pets.
    Otherwise, returns an HTTP 404 Not Found error.

    Args:
        region_tag (str): region_tag abbreviation
        realm_name (str): the slug for the character's realm
        character_name (str): name of character
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
    uri += (
        f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/hunter-pets"
    )

    params = {
        "locale": utils.localize(locale),
        "namespace": namespace(region_tag,  "profile", release)
    }
    return uri, params


@verify_region
def media_summary(
    region_tag: str,
    realm_name: str,
    character_name: str,
    *,
    release: Optional[str] = "retail",
    locale: Optional[str] = None
):
    """Returns a summary of the media assets available for a character (such as an avatar render).

    Args:
        region_tag (str): region_tag abbreviation
        realm_name (str): the slug for the character's realm
        character_name (str): name of character
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
    uri += f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/character-media"

    params = {
        "locale": utils.localize(locale),
        "namespace": namespace(region_tag,  "profile", release)
    }
    return uri, params


@verify_region
def mythic_keystone(
    region_tag: str,
    realm_name: str,
    character_name: str,
    *,
    season_id: Optional[int] = None,
    release: Optional[str] = "retail",
    locale: Optional[str] = None
):
    """Returns the Mythic Keystone profile index for a character,
    or the Mythic Keystone season details for a character.

    Returns a 404 Not Found for characters that have not yet completed
    a Mythic Keystone dungeon for the specified season.

    Args:
        region_tag (str): region_tag abbreviation
        realm_name (str): the slug for the character's realm
        character_name (str): name of character
        season_id (int or None): season id or None (default).  None
            accesses the list of seasons for the current expansion
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
    uri += f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/mythic-keystone-profile"

    if season_id:
        uri += f"/season/{season_id}"

    params = {
        "locale": utils.localize(locale),
        "namespace": namespace(region_tag,  "profile", release)
    }
    return uri, params


@verify_region
def professions_summary(
    region_tag: str,
    realm_name: str,
    character_name: str,
    *,
    release: Optional[str] = "retail",
    locale: Optional[str] = None
):
    """Returns a summary of professions for a character.

    Args:
        region_tag (str): region_tag abbreviation
        realm_name (str): the slug for the character's realm
        character_name (str): name of character
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
    uri += (
        f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/professions"
    )

    params = {
        "locale": utils.localize(locale),
        "namespace": namespace(region_tag,  "profile", release)
    }
    return uri, params


@verify_region
def profile(
    region_tag: str,
    realm_name: str,
    character_name: str,
    *,
    status: bool = False,
    release: Optional[str] = "retail",
    locale: Optional[str] = None
):
    """Returns a profile summary for a character, or the status and a unique ID for a character.

    A should delete information about a character from their application if
    any of the following conditions occur:

        1) an HTTP 404 Not Found error is returned
        2) the is_valid value is false
        3) the returned character ID doesn't match the previously recorded
            value for the character

    The following example illustrates how to use this endpoint:
        1) A requests and stores information about a character, including
        its unique character ID and the timestamp of the request.
        2) After 30 days, the makes a request to the status endpoint to
        verify if the character information is still valid.
        3) If character cannot be found, is not valid, or the characters IDs do
        not match, the removes the information from their application.
        4) If the character is valid and the character IDs match, the retains
        the data for another 30 days.

    Args:
        region_tag (str): region_tag abbreviation
        realm_name (str): the slug for the character's realm
        character_name (str): name of character
        status (bool): flag to request a profile summary (False default) or status (True)
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
    uri += f"{utils.slugify(realm_name)}/{urlquote(character_name.lower())}"

    if status:
        uri += "/status"

    params = {
        "locale": utils.localize(locale),
        "namespace": namespace(region_tag,  "profile", release)
    }
    return uri, params


@verify_region
def pvp(
    region_tag: str,
    realm_name: str,
    character_name: str,
    *,
    pvp_bracket: Optional[str] = None,
    release: Optional[str] = "retail",
    locale: Optional[str] = None
):
    """Returns the PvP bracket statistics for a character, or a PvP summary for a character.

    Args:
        region_tag (str): region_tag abbreviation
        realm_name (str): the slug for the character's realm
        character_name (str): name of character
        pvp_bracket (str or None): '2v2', '3v3', 'rbg', None (default).
            None returns a summary of pvp activity
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    if pvp_bracket and pvp_bracket not in ('2v2', '3v3', '5v5', 'rbg'):
        raise BNetValueError("PvP Bracket not correct:  '2v2', '3v3', '5v5', 'rbg'")

    uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
    uri += f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}"

    if pvp_bracket:
        uri += f"/pvp-bracket/{utils.slugify(pvp_bracket)}"
    else:
        uri += "/pvp-summary"

    params = {
        "locale": utils.localize(locale),
        "namespace": namespace(region_tag,  "profile", release)
    }
    return uri, params


@verify_region
def quests(
    region_tag: str,
    realm_name: str,
    character_name: str,
    *,
    completed: Optional[bool] = False,
    release: Optional[str] = "retail",
    locale: Optional[str] = None,
):
    """Returns a character's active quests as well as a link to the character's completed quests, or
    a list of quests that a character has completed.

    Args:
        region_tag (str): region_tag abbreviation
        realm_name (str): the slug for the character's realm
        character_name (str): name of character
        completed (bool):  To show all quests (False), or to show only
            completed quests (True)
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
    uri += f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/quests"

    if completed:
        uri += f"/completed"

    params = {
        "locale": utils.localize(locale),
        "namespace": namespace(region_tag,  "profile", release)
    }
    return uri, params


@verify_region
def reputations_summary(
    region_tag: str,
    realm_name: str,
    character_name: str,
    *,
    release: Optional[str] = "retail",
    locale: Optional[str] = None
):
    """Returns a summary of a character's reputations.

    Args:
        region_tag (str): region_tag abbreviation
        realm_name (str): the slug for the character's realm
        character_name (str): name of character
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
    uri += (
        f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/reputations"
    )

    params = {
        "locale": utils.localize(locale),
        "namespace": namespace(region_tag,  "profile", release)
    }
    return uri, params


@verify_region
def soulbinds(
    region_tag: str,
    realm_name: str,
    character_name: str,
    *,
    release: Optional[str] = "retail",
    locale: Optional[str] = None
):
    """Returns a character's soulbinds.

    Args:
        region_tag (str): region_tag abbreviation
        realm_name (str): the slug for the character's realm
        character_name (str): name of character
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
    uri += f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/soulbinds"

    params = {
        "locale": utils.localize(locale),
        "namespace": namespace(region_tag,  "profile", release)
    }
    return uri, params


@verify_region
def specializations_summary(
    region_tag: str,
    realm_name: str,
    character_name: str,
    *,
    release: Optional[str] = "retail",
    locale: Optional[str] = None
):
    """Returns a summary of a character's specializations.

    Args:
        region_tag (str): region_tag abbreviation
        realm_name (str): the slug for the character's realm
        character_name (str): name of character
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
    uri += f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/specializations"

    params = {
        "locale": utils.localize(locale),
        "namespace": namespace(region_tag,  "profile", release)
    }
    return uri, params


@verify_region
def statistics_summary(
    region_tag: str,
    realm_name: str,
    character_name: str,
    *,
    release: Optional[str] = "retail",
    locale: Optional[str] = None
):
    """Returns a statistics summary for a character.

    Args:
        region_tag (str): region_tag abbreviation
        realm_name (str): the slug for the character's realm
        character_name (str): name of character
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
    uri += f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/statistics"

    params = {
        "locale": utils.localize(locale),
        "namespace": namespace(region_tag,  "profile", release)
    }
    return uri, params


@verify_region
def title_summary(
    region_tag: str,
    realm_name: str,
    character_name: str,
    *,
    release: Optional[str] = "retail",
    locale: Optional[str] = None
):
    """Returns a summary of titles a character has obtained.

    Args:
        region_tag (str): region_tag abbreviation
        realm_name (str): the slug for the character's realm
        character_name (str): name of character
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
    uri += f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/titles"

    params = {
        "locale": utils.localize(locale),
        "namespace": namespace(region_tag,  "profile", release)
    }
    return uri, params


@verify_region
def guild(
    region_tag: str,
    realm_name: str,
    guild_name: str,
    *,
    category: Optional[str] = None,
    release: Optional[str] = "retail",
    locale: Optional[str] = None
):
    """Returns a single guild by its name and realm, achievements, activity, or roster

    Args:
        region_tag (str): region_tag abbreviation
        realm_name (str): the name of the guild's realm
        guild_name (str): the name of the guild
        category (str): category of guild data to retrieve
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/guild/"
    uri += f"{utils.slugify(realm_name)}/{utils.slugify(guild_name)}"

    if category:
        if category in ("activity", "achievements", "roster"):
            uri += f"/{category}"
        else:
            raise BNetValueError("Category Invalid: current choices activity, achievement and roster")

    params = {
        "locale": utils.localize(locale),
        "namespace": namespace(region_tag, "profile", release)
    }
    return uri, params



"""Playable Race, Class, and Specializations for World of Warcraft

A class in World of Warcraft, a job the character does.

Author:
David "Gahd" Couples <gahdania@gahd.online>
"""
from battlenet_client.wow.base import WoWBase, WoWMedia
from battlenet_client.cache.cache import Cache
from battlenet_client.exceptions import BNetValueError
from battlenet_client.wow.game_data import playable_race, playable_class, playable_class_media
from battlenet_client.wow.game_data import playable_spec, playable_spec_media, pvp_talent, power_type
from battlenet_client.utils import localize, slugify

import json
from oic import oic
from requests_oauthlib import OAuth2Session
from typing import Optional, Union, List, Dict


class PowerType(WoWBase):
    """Playable Specialization
    """
    def __init__(self, client: Union[OAuth2Session, oic.Client], region_tag: str, power_type_id: int, *,
                 release: Optional[str] = 'retail', locale: Optional[str] = 'enus', cache: Optional[Cache] = None):
        """Initializes :obj:`PowerType`

        Args:
            client (:obj:`OAuth2Session` or :obj:`oic.Client`): the client to communicate to the API
            region_tag (str): the region abbreviation
            class_id (int): the ID of the class

        Keywords Args:
            release (str): release of the game
            locale (str): localization to use
            cache (:obj:`Cache`): the cache to use for storage
        """
        (url, params) = power_type(region_tag, power_type_id, release=release)
        api_data = self.get(client, release, 'power-type', url, params, object_id=power_type_id, cache=cache)

        super().__init__(region_tag, api_data['id'], release=release, locale=locale)

        self._names = api_data['name']

    def __str__(self):
        return f"{self.get_name(self.locale) if self.locale else self.get_name('en_US')} ({self.region_tag.upper()})"

    def __repr__(self):
        return f"{self.get_name(self.locale) if self.locale else self.get_name('en_US')} ({self.region_tag.upper()})"

    def get_name(self, locale: Optional[str] = None):
        """Returns the name of the faction by `locale`

        Args:
            locale (str: optional): locale to reference.  Defaults to this :obj:`PlayableRace` locale

        Returns:
            str: the localized name

        Note:
            Some languages have nouns that have gender.
        """
        return self._localized_data('_names', locale)


class PlayableSpecialization(WoWBase):
    """Playable Specialization

    Attributes:
        icon (:obj:`Image`): stores info about the icon
        playable_class (:obj:`PlayableClass`): the playable class that uses this specialization
        talent_tree_url (str): URL for the talent tree
        pvp_talents (dict): the Person Versus Person (PvP) talents
        icon (:obj:`WoWMedia): icon of the spec
    """
    def __init__(self, client: Union[OAuth2Session, oic.Client], region_tag: str, spec_id: int, *,
                 release: Optional[str] = 'retail', locale: Optional[str] = 'enus', cache: Optional[Cache] = None):
        """Initializes :obj:`PlayableSpecialization`

        Args:
            client (:obj:`OAuth2Session` or :obj:`oic.Client`): the client to communicate to the API
            region_tag (str): the region abbreviation
            class_id (int): the ID of the class

        Keywords Args:
            release (str): release of the game
            locale (str): localization to use
            cache (:obj:`Cache`): the cache to use for storage
        """
        (url, params) = playable_spec(region_tag, spec_id, release=release)
        api_data = self.get(client, release, 'playable-specialization', url, params, object_id=spec_id,
                            cache=cache)

        super().__init__(region_tag, api_data['id'], release=release, locale=locale)

        self._names = api_data['name']
        self._descriptions = {key: value for key, value in api_data['gender_description'].items()}
        self._roles = api_data['role']['name']
        self._primary_stat = api_data['primary_stat_type']['name']

        if api_data['playable_class']['id'] > 0:
            self.playable_class = api_data['playable_class']['id']

        if 'spec_talent_tree' in api_data.keys():
            self.talent_tree_url = api_data['spec_talent_tree']['key']['href']

        if 'pvp_talents' in api_data.keys():
            self.pvp_talents = api_data['pvp_talents']

        (url, params) = playable_spec_media(region_tag, spec_id, release=release)
        api_data = self.get(client, release, 'playable-specialization-media', url, params, object_id=spec_id,
                            cache=cache)
        self.icon = WoWMedia(api_data)

    def __str__(self):
        return f"{self.get_name(self.locale) if self.locale else self.get_name('en_US')} ({self.region_tag.upper()})"

    def __repr__(self):
        return f"{self.get_name(self.locale) if self.locale else self.get_name('en_US')} ({self.region_tag.upper()})"

    def get_description(self, gender: Optional[str] = 'neutral', locale: Optional[str] = None):
        """Returns the description of the specialization by gender and locale.

        Args:
            gender (str: optional): gender to reference.  Defaults to `female`
            locale (str: optional): locale to reference.  Defaults to this :obj:`PlayableSpecialization` locale

        Returns:
            str: the localized name for the given gender

        Note:
            Some languages have nouns that have gender.
        """
        return self._gender_data('_descriptions', gender, locale)

    def get_name(self, locale: Optional[str] = None):
        """Returns the name of the faction by `locale`

        Args:
            locale (str: optional): locale to reference.  Defaults to this :obj:`PlayableRace` locale

        Returns:
            str: the localized name for the given gender

        Note:
            Some languages have nouns that have gender.
        """
        return self._localized_data('_names', locale)

    def get_role(self, locale: Optional[str] = None):
        """Returns the name of the faction by `locale`

        Args:
            locale (str: optional): locale to reference.  Defaults to this :obj:`PlayableRace` locale

        Returns:
            str: the localized name for the given gender

        Note:
            Some languages have nouns that have gender.
        """
        return self._localized_data('_roles', locale)

    def get_primary_stat(self, locale: Optional[str] = None):
        """Returns the name of the faction by `locale`

        Args:
            locale (str: optional): locale to reference.  Defaults to this :obj:`PlayableRace` locale

        Returns:
            str: the localized name for the given gender

        Note:
            Some languages have nouns that have gender.
        """
        return self._localized_data('_primary_stats', locale)


class PlayableClass(WoWBase):
    """Defines the playable class.


    Attributes:
        class_id (int): playable class ID
        names (dict of str): localized names of the class
        power_type (:obj:`PowerType`):  power type used for class, IE mana, rage, fury, focus
        icon (:obj:`WoWMedia`):  the class specific icon
        specializations (list of :obj:`PlayableSpecialization`): the available specializations for the class
        additional_power_type (:obj:`PowerType`): the additional power type used by the class, IE maelstrom stacks
        races (list of :obj:`PlayableRace`): the races that can be the class
    """
    def __init__(self, client: Union[OAuth2Session, oic.Client], region_tag: str, class_id: int, *,
                 release: Optional[str] = 'retail', locale: Optional[str] = 'enus', cache: Optional[Cache] = None):
        """Initializes the :obj:`PlayableClass`

        Args:
            client (:obj:`OAuth2Session` or :obj:`oic.Client`): the client to communicate to the API
            region_tag (str): the region abbreviation
            class_id (int): the ID of the class

        Keywords Args:
            release (str): release of the game
            locale (str): localization to use
            cache (:obj:`Cache`): the cache to use for storage
        """
        (url, params) = playable_class(region_tag, class_id, release=release)
        api_data = self.get(client, release, 'playable-class', url, params, object_id=class_id,
                            cache=cache)

        super().__init__(region_tag, api_data['id'], release=release, locale=locale)

        self._names = {'female': api_data['gender_name']['female'], 'male': api_data['gender_name']['male'],
                       'neutral': api_data['name']}
        self.power_type = api_data['power_type']['id']

        self._specializations = [spec['id'] for spec in api_data['specializations']]
        self._additional_power_types = [power['id'] for power in api_data['additional_power_types']]
        self._races = [race['id'] for race in api_data['playable_races']]

        (url, params) = playable_class_media(region_tag, self.id, release=release)
        api_data = self.get(client, release, 'playable-class-media', url, params, object_id=self.id,
                            cache=cache)

        self.icon = WoWMedia(api_data)

    def __str__(self):
        name = self.get_name(locale=self.locale) if self.locale else self.get_name(locale='en_US')
        return f"{name} ({self.region_tag.upper()})"

    def __repr__(self):
        name = self.get_name(locale=self.locale) if self.locale else self.get_name(locale='en_US')
        return f"{name} ({self.region_tag.upper()})"

    def get_name(self, gender: Optional[str] = 'neutral', locale: Optional[str] = None) -> str:
        """Returns the name of the race by gender and locale.

        Args:
            gender (str: optional): gender to reference.  Defaults to `female`
            locale (str: optional): locale to reference.  Defaults to this :obj:`PlayableRace` locale

        Returns:
            str: the localized name for the given gender

        Note:
            Some languages have nouns that have gender.
        """
        return self._gender_data('_names', gender, locale)

    def is_additional_power_type_usable(self, power_type_id: int) -> bool:
        """ Returns the power type of

        Args:
            power_type_id (int): ID of the power type

        Returns:
            :obj:`Realm`: realm identified by its ID or None if not found
        """
        return power_type_id in self._additional_power_types

    def is_race_usable(self, race_id: int) -> bool:
        """ Returns if the race identified by `race_id` is used by :obj:`PlayableClass`

        Args:
            race_id (int): ID of the power type

        Returns:
            :obj:`PlayableRace`: realm identified by its ID or None if not found
        """
        return race_id in self._races

    def is_spec_usable(self, spec_id: int) -> bool:
        """ Returns if the race identified by `race_id` is used by :obj:`PlayableClass`

       Args:
           spec_id (int): ID of the specialization

       Returns:
           :obj:`PlayableRace`: realm identified by its ID or None if not found
       """
        return spec_id in self._specializations


class PlayableRace(WoWBase):
    """Playable Race

    Attributes:
        id (int): the playable race's ID
        faction (str): the faction the playable race is
        female_name (str): the feminine version of the name in localizations that support noun gender (romance lang)
        male_name (str): the masculine version of the name in localizations that support noun gender (romance lang)
        name (str): the neuter form
        is_selectable (bool): flag where the playable race selectable on the character select screen
        is_allied_race (bool): flag where the playable race is an allied race
    """
    def __init__(self, client: Union[OAuth2Session, oic.Client], region_tag: str, race_id: int, *,
                 release: Optional[str] = 'retail', locale: Optional[str] = None, cache: Optional[Cache] = None):
        """Initializes this instance of :obj:`PlayableRace`

        Args:
            client (object: `Requests`): OIDC/OAuth v2 compatible client
            region_tag (str): region abbreviation for use with the APIs
            race_id (str, int, optional):  race id to use
            release (str, optional): the release to use. Defaults to "retail"
            locale (str, optional): the localization to use from the API.  None uses region_tag's default locale

        Note:
            Playable Race does not have a media type. See :obj:`Character`
        """
        (url, params) = playable_race(region_tag, race_id, release=release)
        api_data = self.get(client, release, 'playable-race', url, params, object_id=race_id, cache=cache)

        super().__init__(region_tag, api_data['id'], release=release, locale=locale)
        self._factions = api_data['faction']['name']
        self._names = {key: value for key, value in api_data['gender_name'].items()}
        self._names.update({'neutral': api_data['name']})
        self.is_selectable = bool(api_data['is_selectable'])
        self.is_allied_race = bool(api_data['is_allied_race'])
        self._available_classes = [cls['id'] for cls in api_data['playable_classes']]

    def __str__(self) -> str:
        name = self.get_name(locale=self.locale) if self.locale else self.get_name(locale='en_US')
        return f"{name} ({self.region_tag.upper()})"

    def __repr__(self) -> str:
        name = self.get_name(locale=self.locale) if self.locale else self.get_name(locale='en_US')
        return f"{name} ({self.region_tag.upper()})"

    def get_name(self, gender: Optional[str] = 'neutral', locale: Optional[str] = None) -> str:
        """Returns the name of the race by gender and locale.

        Args:
            gender (str: optional): gender to reference.  Defaults to `female`
            locale (str: optional): locale to reference.  Defaults to this :obj:`PlayableRace` locale

        Returns:
            str: the localized name for the given gender

        Note:
            Some languages have nouns that have gender.
        """
        return self._gender_data('_names', gender, locale)

    def get_faction(self, locale: Optional[str] = None) -> str:
        """Returns the name of the faction by `locale`

        Args:
            locale (str: optional): locale to reference.  Defaults to this :obj:`PlayableRace` locale

        Returns:
            str: the localized name for the given gender

        Note:
            Some languages have nouns that have gender.
        """
        return self._localized_data('_factions', locale)

    def is_playable_class_available_to_race(self, class_id: int) -> bool:
        """ Returns if the playable class identified by `class_id` available to :obj:`PlayableRace`

        Args:
            class_id (int): ID of the player class

        Returns:
            :obj:`Realm`: realm identified by its ID or None if not found
        """
        return class_id in self._available_classes



"""Defines the Mythic Keystone Classes

Disclaimer:
All rights reserved, Blizzard is the intellectual property owner of WoW and WoW Classic
and any data pertaining thereto

Author:
David "Gahd" Couples <gahdania@gahd.online>
"""
from datetime import datetime

from battlenet_client.wow.game_data import mythic_keystone_affix, mythic_keystone_dungeon, mythic_keystone_index
from battlenet_client.wow.game_data import mythic_keystone_season, mythic_keystone_leaderboard


class MythicKeystoneAffix:
    """Index of Mythic Keystone Affixes

    Args:
        region_tag (str): region desired for the request

    Keyword Args:
         release (str, optional): the release to use.
         locale (str, optional): the localization to use from the API, The default of None means all localizations
         scope (list of str, optional): the scope or scopes to use during the endpoints that require the
             Web Application Flow
         redirect_uri (str, optional): the URI to return after a successful authentication between the user and Blizzard
         client_id (str): the client ID from the developer portal
         client_secret (str): the client secret from the developer portal

    Attributes:
        locale (str): localization
        affixes (list): the list of the quest areas
    """

    def __init__(self, client: Union[OAuth2Session, OicClient], region_tag: str, affix_id: Optional[str, int] = None, *,
                 release: Optional[str] = None, locale: Optional[str] = None):

        (url, params) = mythic_keystone_affix(region_tag, affix_id, release=release, locale=locale)
        api_data = client.get(url, params=params)

        if not affix_id:
            self.affixes = api_data['affixes']
        else:
            self.affix_id = api_data['id']
            self.name = api_data['name']
            self.description = api_data['description']
            # media = self._client.mythic_keystone_affix_media(affix_id)
            # self.media = Image.open(self._client.get(media['assets'][0]['value']))


class MythicKeystoneDungeon:
    """Index of Mythic Keystone Dungeons

    Args:
        region_tag (str): region desired for the request

    Keyword Args:
         release (str, optional): the release to use.
         locale (str, optional): the localization to use from the API, The default of None means all localizations
         scope (list of str, optional): the scope or scopes to use during the endpoints that require the
             Web Application Flow
         redirect_uri (str, optional): the URI to return after a successful authentication between the user and Blizzard
         client_id (str): the client ID from the developer portal
         client_secret (str): the client secret from the developer portal

    Attributes:
        locale (str): localization
        dungeons (list): the list of the quest areas
    """
    def __init__(self, client: Union[OAuth2Session, OicClient], region_tag: str, dungeon_id: Optional[str, int] = None, *,
                 release: Optional[str] = None, locale: Optional[str] = None):

        (url, params) = mythic_keystone_dungeon(region_tag, dungeon_id, release=release, locale=locale)
            
        api_data = client.get(url, params=params)

        if not dungeon_id:
            self.dungeons = api_data['dungeons']
        else:

            self.object_id = api_data['id']
            self.name = api_data['name']
            self.map = api_data['map']
            self.zone_slug = api_data['zone']['slug']
            self.breakpoints = {chest['upgrade_level']: chest['qualifying_duration']
                                for chest in api_data['keystone_upgrades']}
            self.dungeon = api_data['dungeon']


class MythicKeystoneIndex:
    """Index of Mythic Keystones

    Args:
        region_tag (str): region desired for the request

    Keyword Args:
         release (str, optional): the release to use.
         locale (str, optional): the localization to use from the API, The default of None means all localizations
         scope (list of str, optional): the scope or scopes to use during the endpoints that require the
             Web Application Flow
         redirect_uri (str, optional): the URI to return after a successful authentication between the user and Blizzard
         client_id (str): the client ID from the developer portal
         client_secret (str): the client secret from the developer portal

    Attributes:
        locale (str): localization
        seasons (list): the seasons for mythic keystones
        dungeons (list): the dungeons for the mythic keystones
    """
    def __init__(self, client: Union[OAuth2Session, OicClient], region_tag: str, *, release: Optional[str] = None, locale: Optional[str] = None):

        (url, params) = mythic_keystone_index(region_tag, release=release, locale=locale)

        api_data = client.get(url, params=params)
        seasons = client.get(api_data['seasons']['href'])
        dungeons = client.get(api_data['dungeons']['href'])

        for season in seasons['seasons']:
            api_data = client.get(season['key']['href'])
            self.seasons.append(MythicKeystoneSeason(client, region_tag, api_data['id'], release=release,
                                                     locale=locale))

        for dungeon in dungeons['dungeons']:
            api_data = client.get(dungeon['key']['href'])
            self.dungeons.append(MythicKeystoneDungeon(client, region_tag, api_data['id'], release=release,
                                                       locale=locale))


class MythicKeystonePeriod:
    """Index of Mythic Keystone Periods

    Args:
        region_tag (str): region desired for the request

    Keyword Args:
         release (str, optional): the release to use.
         locale (str, optional): the localization to use from the API, The default of None means all localizations
         scope (list of str, optional): the scope or scopes to use during the endpoints that require the
             Web Application Flow
         redirect_uri (str, optional): the URI to return after a successful authentication between the user and Blizzard
         client_id (str): the client ID from the developer portal
         client_secret (str): the client secret from the developer portal

    Attributes:
        locale (str): localization
        periods (list): the list of the dungeons for mythic keystones
    """

    def __init__(self, client: Union[OAuth2Session, OicClient], region_tag: str, period_id: Optional[str, int] = None, *,
                 release: Optional[str] = None, locale: Optional[str] = None):

        (url, params) = mythic_keystone_period(region_tag, period_id, release=release, locale=locale)

        api_data = client.get(url, params=params)

        if not period_id:
            self.periods = api_data['periods']
        else:
            self.object_id = api_data['id']
            self.start = daapi_data['start_timestamp']
            self.end = api_data['end_timestamp']


class MythicKeystoneSeason:
    """Index of Mythic Keystone Season

    Args:
        region_tag (str): region desired for the request

    Keyword Args:
         release (str, optional): the release to use.
         locale (str, optional): the localization to use from the API, The default of None means all localizations
         scope (list of str, optional): the scope or scopes to use during the endpoints that require the
             Web Application Flow
         redirect_uri (str, optional): the URI to return after a successful authentication between the user and Blizzard
         client_id (str): the client ID from the developer portal
         client_secret (str): the client secret from the developer portal

    Attributes:
        locale (str): localization
        seasons (list): the list of the seasons for mythic keystones
        current_season (dict): the
    """

    def __init__(self, client: Union[OAuth2Session, OicClient], region_tag: str, season_id: Optional[str, int] = None, *,
                 release: Optional[str] = None, locale: Optional[str] = None):

        (url, params) = mythic_keystone_season(region_tag, season_id, release=release, locale=locale)
        api_data = client.get(url, params=params)

        if not season_id:
            self.seasons = api_data['seasons']
            self.current_season = api_data['current_season']
        else:
            self.season_id = api_data['id']
            self.start = datetime.utcfromtimestamp(api_data['start_timestamp'] / 1000)
            self.end = datetime.utcfromtimestamp(api_data['end_timestamp'] / 1000)
            self.periods = api_data['periods']


class MythicKeystoneLeaderBoard:
    """Index of the Mythic Keystone Leaderboards

    Args:
        region_tag (str): region desired for the request

    Keyword Args:
         release (str, optional): the release to use.
         locale (str, optional): the localization to use from the API, The default of None means all localizations
         scope (list of str, optional): the scope or scopes to use during the endpoints that require the
             Web Application Flow
         redirect_uri (str, optional): the URI to return after a successful authentication between the user and Blizzard
         client_id (str): the client ID from the developer portal
         client_secret (str): the client secret from the developer portal

    Attributes:
        locale (str): localization
        leaderboards (list): the list of the quest areas

    """

    def __init__(self, client: Union[OAuth2Session, OicClient], region_tag: str, connected_realm_id: Optional[str, int] = None, *,
                 release: Optional[str] = None, locale: Optional[str] = None):

        (url, params) = mythic_keystone_leaderboard(region_tag, connected_realm_id, release=release, locale=locale)
        api_data = client.get(url, params=params)

        if not connected_realm_id:
            self.leaderboards = api_data['leaderboards']

        else:
            self.map = api_data['map']
            self.period = int(api_data['period'])
            self.period_start_timestamp = int(api_data['period_start_timestamp'])
            self.period_end_timestamp = int(api_data['period_end_timestamp'])
            self.connected_realm = api_data['connected_realm']['href']
            self.leading_groups = api_data['leading_groups']



import pytest
from battlenet_client import utils
from battlenet_client.exceptions import BNetValueError


@pytest.mark.parametrize(
    "value",
    [(1723, 40, 32), [1723, 40, 32], {"gold": 1723, "silver": 40, "copper": 32}],
)
def test_currency_convertor_invalid_type(value):
    with pytest.raises(TypeError):
        utils.currency_convertor(value)


def test_currency_convertor_negative_number():
    with pytest.raises(BNetValueError):
        utils.currency_convertor(-43_23_11)


@pytest.mark.parametrize("value", ["17234032", 17234032])
def test_currency_convertor_str(value):
    value = utils.currency_convertor(value)
    assert isinstance(value, tuple)
    assert value == (1723, 40, 32)


def test_slugify_with_int():
    result = utils.slugify(81)
    assert isinstance(result, int)
    assert result == 81


def test_slugify_with_tuple():
    with pytest.raises(AttributeError):
        # noinspection PyTypeChecker
        utils.slugify((1, 3, 4))


def test_slugify_with_str():
    result = utils.slugify("We'll Be Back")
    assert isinstance(result, str)
    assert result == 'well-be-back'


def test_localize_with_none():
    result = utils.localize(None)
    assert result is None


def test_localize_with_int():
    with pytest.raises(BNetValueError):
        # noinspection PyTypeChecker
        utils.localize(81)


def test_localize_with_aaus():
    with pytest.raises(BNetValueError):
        utils.localize('aaUS')


def test_localize_with_enaa():
    with pytest.raises(BNetValueError):
        utils.localize('enaa')


def test_localize_with_enus():
    result = utils.localize('enUS')
    assert isinstance(result, str)
    assert result == 'en_US'


def test_api_host_cn():
    result = utils.api_host('CN')
    assert result == "https://gateway.battlenet.com.cn"


def test_api_host_us():
    result = utils.api_host('US')
    assert result == "https://us.api.blizzard.com"


def test_auth_host_cn():
    result = utils.auth_host('CN')
    assert result == "https://www.battlenet.com.cn"


def test_auth_host_us():
    result = utils.auth_host('US')
    assert result == "https://us.battle.net"


def test_auth_host_tw():
    result = utils.auth_host('TW')
    assert result == "https://apac.battle.net"


def test_render_host_cn():
    result = utils.render_host('CN')
    assert result == "https://render.worldofwarcraft.com.cn"


def test_render_host_us():
    result = utils.render_host('US')
    assert result == "https://render-us.worldofwarcraft.com"





"""World of Warcraft Guild and Banner

Disclaimer:
All rights reserved, Blizzard is the intellectual property owner of WoW and WoW Classic
and any api_data pertaining thereto

Author:
David "Gahd" Couples <gahdania@gahd.online>
"""
import os
import pendulum
from PIL import Image, ImageChops, ImageDraw, ImageFont

from wow_api.clients.client import WoWClient
from .realms import Realm
from battlenet_client.wow.regions import Region
from .characters import Character


class Guild:
    """World of Warcraft Guild

    Args:
        region_tag (str or :obj:`Region`): region abbreviation for use with the APIs
        realm (int, str, :obj:`Realm`): the ID or name of the realm
        name (str): the name or slug of the guild

    Keyword Args:
        release (str, optional): the release to use.  Currently 'classic' only.  The default of None, indicates to use
            the retail data
        locale (str, optional): the localization to use from the API, The default of None means all localizations
        scope (list of str, optional): the scope or scopes to use during the endpoints that require the
            Web Application Flow
        redirect_uri (str, optional): the URI to return after a successful authentication between the user and Blizzard
        client_id (str): the client ID from the developer portal
        client_secret (str): the client secret from the developer portal
    """
    def __init__(self, client: Union[OAuth2Session, OicClient], region_tag: str, realm: Union[str, int], name: str, *,
                 release: Optional[str] = None, locale: Optoinal[str] = None):

        self.region = Region(client, region_tag, release=release, locale=locale)
        self.realm = Realm(client, region_tag, api_data['realm']['id'], locale=locale, release=release)

        self.object_id = api_data['id']
        self.name = api_data['name']
        self.faction = api_data['faction']

        self.member_count = int(api_data['member_count'])
        self.achievement_points = int(api_data['achievement_points'])
        self._achievements_url = api_data['achievements']['href']
        self.crest = api_data['crest']
        
        self.created = pendulum.from_timestamp(api_data['created_timestamp'] / 1000, tz=self.realm.timezone)
        self.roster_list = self._client.guild_roster(self.realm.slug, self.slug)
        self.achievements = self._client.guild_achievements(self.realm.slug, self.slug)
        self.activities = self._client.guild_activities(self.realm.slug, self.slug)

        self.level = None if 'level' not in api_data.keys() else api_data['level']

    def __str__(self):
        return f"{self.name}-{self.realm.name} ({self.realm.region.tag.upper()})"

    @property
    def roster(self):
        return [Character('us', character['character']['realm']['slug'], character['character']['name'],
                locale=self.locale) for character in self.roster_list['members']]


class GuildBanner:

    def __init__(self, region, guild, *, release=None, locale=None, scope=None, redirect_uri=None,
                 client=None, client_id=None, client_secret=None):

        if client:
            self._client = client
        else:
            self._client = WoWClient(region, locale=locale, release=release, scope=scope, redirect_uri=redirect_uri,
                                     client_id=client_id, client_secret=client_secret)

        if locale:
            self.locale = locale
        else:
            self.locale = self._client.locale

        if isinstance(region, Region):
            self.region = region
        else:
            self.region = Region(region, locale=locale, release=release, scope=scope, redirect_uri=redirect_uri,
                                 client=self._client)

        self.guild_id = int(guild.id)
        self.faction = guild.faction
        self.level = guild.level

        self.emblem = dict(id=None, color=dict(r=None, g=None, b=None, a=None))
        self.border = dict(id=None, color=dict(r=None, g=None, b=None, a=None))
        self.banner = dict(r=None, g=None, b=None, a=None)

        self.emblem['id'] = guild.crest['emblem']['id']
        for key, value in guild.crest['emblem']['color']['rgba'].items():
            if key == 'a':
                value = int(value * 255)

            self.emblem['color'][key] = value

        self.border['id'] = guild.crest['border']['id']
        for key, value in guild.crest['border']['color']['rgba'].items():
            if key == 'a':
                value = int(value * 255)
            self.border['color'][key] = value

        for key, value in guild.crest['background']['color']['rgba'].items():
            if key == 'a':
                value = int(value * 255)
            self.banner[key] = value

    def show_crest(self, show_ring=True, show_level=False, width=215, path=None):

        if path and isinstance(path, str):
            path = os.path.dirname(path)

        if not path:
            path = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'crest')

        crest_file = os.path.join(path, 'cache', f"{self.region.id}_{self.guild_id}.png")

        try:
            crest = Image.open(crest_file)
        except FileNotFoundError:
            crest = self.create_crest(show_ring=show_ring, show_level=show_level, width=width, path=path)
        except AttributeError:
            crest = self.create_crest(show_ring=show_ring, show_level=show_level, width=width, path=path)

        return crest

    def delete_crest(self, path=None):

        if path and isinstance(path, str):
            path = os.path.dirname(path)

        if not path:
            path = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'crest')

        crest = os.path.join(path, 'cache', f"{self.region.id}_{self.guild_id}.png")

        try:
            os.unlink(crest)
        except OSError as error:
            print(error)
        else:
            return None

    def create_crest(self, show_ring=True, show_level=False, width=215, path=None):
        """Creates the guild emblem

        Args:
            show_ring (bool): show the faction (default) if True, else do not (False)
            show_level (bool): True to show the guild level, False otherwise (default)
            width (int): the overall width of the guild emblem
            path (:obj:`os.path`): the path to the segments

        Returns:
            (:obj:`Image`): the completed image of the emblem
        """

        def _blend(base, color):
            """builds the layer to be the next overlay of the images

            Args:
                base (:obj:`Image`): the image to use as the base of the overlay, ie the emblem, border, or flag
                    background converted to RGBA format
                color (dict): the color in red (r), green (g), blue (b), and alpha (a)

            Returns:
                (:obj:`Image`): completed layer to overlay onto the temporary image.
            """

            swatch = Image.new('RGB', base.size, (color['r'], color['g'], color['b']))
            # alpha = base.getchannel('A').convert('RGB')
            # need to find right modification to emulate a merge based on luminosity
            image = ImageChops.overlay(swatch, base.convert('RGB')).convert('RGBA')
            image.putalpha(base.getchannel('A'))
            return image

        x_offset = 20
        y_offset = 23
        height = int(width * 230 / 215)
        scale = width / 215

        if path and isinstance(path, str):
            path = os.path.dirname(path)

        if not path:
            path = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'crest')

        static = os.path.join(path, 'static')
        save_path = os.path.join(path, 'cache', f"{self.region.id}_{self.guild_id}.png")

        # load the images
        output = Image.new('RGBA', (int(width), int(height)), (0, 0, 0, 0))
        ring = Image.open(os.path.join(static, f"ring-{self.faction['type'].lower()}.png")).convert('RGBA')
        shadow = Image.open(os.path.join(static, 'shadow_00.png')).convert('RGBA')
        flag = Image.open(os.path.join(static, 'bg_00.png')).convert('RGBA')
        emblem_data = self._client.guild_crest_emblem_media(self.emblem['id'])
        emblem = Image.open(self._client.get(emblem_data['assets'][0]['value']))
        border_data = self._client.guild_crest_border_media(self.border['id'])
        border = Image.open(self._client.get(border_data['assets'][0]['value'])).convert('RGBA')
        hooks = Image.open(os.path.join(static, 'hooks.png')).convert('RGBA')

        flag = _blend(flag, self.banner)

        overlay = Image.open(os.path.join(static, 'overlay_00.png')).convert('RGBA')

        emblem = _blend(emblem, self.emblem['color'])

        border = _blend(border, self.border['color'])

        # load the faction ring, and lowest shadow
        if show_ring:
            output.alpha_composite(ring)
            output.alpha_composite(shadow, (x_offset, y_offset))

        output.alpha_composite(flag, (x_offset, y_offset))
        output.alpha_composite(emblem, (x_offset + 17, y_offset + 22))
        output.alpha_composite(border, (x_offset + 13, y_offset + 14))
        output.alpha_composite(overlay, (x_offset, y_offset + 2))
        output.alpha_composite(hooks, (x_offset - 2, y_offset))

        if show_level and self.level:
            level_output = Image.new('RGBA', output.size, (0, 0, 0, 0))
            draw = ImageDraw.Draw(level_output)
            font = ImageFont.truetype(os.path.join(static, 'CantataOne-Regular.ttf'), 80)

            text_size = font.getsize(str(self.level))

            if show_ring:
                level_output.alpha_composite(ring)

            # add the flag shadow layer
            level_output.alpha_composite(shadow, (x_offset, y_offset))

            # add the flag
            level_output.alpha_composite(flag, (x_offset, y_offset))

            offsets = {
                'x1': int((ring.width - text_size[0]) / 2),
                'y1': int((ring.height - text_size[1]) / 2),
                'x2': int((ring.width - text_size[0]) / 2) + text_size[0],
                'y2': int((ring.height - text_size[1]) / 2) + text_size[1],
                'tx': font.getoffset(str(self.level))[0],
                'ty': font.getoffset(str(self.level))[1],
            }

            draw.text((offsets['x1'] - offsets['tx'] / 2, offsets['y1'] - offsets['ty'] / 2), str(self.level),
                      font=font,
                      fill=(self.emblem['color']['r'], self.emblem['color']['g'], self.emblem['color']['b'],
                            self.emblem['color']['a']),
                      outline=(self.border['color']['r'], self.border['color']['g'], self.border['color']['b'],
                               self.border['color']['a']),
                      width=5)

            # add the border
            level_output.alpha_composite(border, (x_offset + 13, y_offset + 15))
            level_output.alpha_composite(overlay, (x_offset, y_offset + 2))
            level_output.alpha_composite(hooks, (x_offset - 2, y_offset))
            level_output = level_output.resize((int(215 / 3), int(230 / 3)), resample=Image.BICUBIC)
            output.alpha_composite(level_output, (215 - level_output.width, 230 - level_output.height))

        if scale != 1:
            output.resize((output.width * scale, output.height * scale), resample=Image.BICUBIC)
        print(save_path)
        output.save(save_path)



"""World of Warcraft Regions

Classes:
    Region

Author:
    David "Gahd" Couples (gahdania@gahd.io)

License:
    See LICENSE file
"""


__version__ = "1.0.1"


from battlenet_client.wow.base import WoWBase
from battlenet_client.cache.cache import Cache
from battlenet_client.exceptions import BNetValueError, BNetRegionError
from battlenet_client.wow.game_data import region

import json
from oic import oic
from requests_oauthlib import OAuth2Session
from typing import Optional, Union, List, Dict


class Region(WoWBase):
    """World of Warcraft Region

    Attributes:
        tag (str): abbreviation for the :obj:`Region`
        region_id (int): numeric representation for the :obj:`Region`
    """
    
    def __init__(self, client: Union[OAuth2Session, oic.Client], region_tag: str, region_id: Optional[int] = None, *,
                 release: Optional[str] = 'retail', locale: Optional[str] = None, cache: Optional[Cache] = None):
        """Initializes an instance of :obj:`Region`

        Args:
            client (:obj:`OAuth2Session`, :obj:`oic.Client`): used with communications to the API
            region_tag (str): region client is requesting data for
            region_id (int): region ID for the requested region (not to be confused with `region_tag`)

        Keyword Args:
            release (str, optional): the release of the game to lookup, defaults to `retail`
            locale (str, optional): localization to use, defaults to all localizations
            cache (:obj:`Cache`, optional): what cache to use

        Raises:
            BNetValueError: when the API returns a 404 error
        """

        if region_id:
            (url, params) = region(self.region_tag, region_id, release=self.release)
            api_data = self.get(client, release, 'region', url, params, object_id=region_id, cache=cache)

        else:
            # dynamically determine region for `region_tag`
            # api docs call this a `Region Index`
            (url, params) = region(self.region_tag, release=self.release)
            region_index = self.get(client, release, 'region', url, params, cache=cache)

            # bypassing self.get here since
            # URL retrieved above has additional query parameters included
            region_data = client.get(region_index['regions'][0]['href'])
            if region_data.status_code == 404:
                raise BNetValueError(raw_data.text)

            api_data = json.loads(region_data.content)

            if cache:
                cache.submit(self.region_tag, self.release, 'region', region_data, api_data['id'])

        super().__init__(region_tag, api_data['id'], release=release, locale=locale)

        self._names = api_data['name']
        self.tag = api_data['tag']

    def __str__(self):
        return f"{self.name(self.locale) if self.locale else self.name('enus')} ({self.region_tag.upper()})"

    def __repr__(self):
        return f"{self.name(self.locale) if self.locale else self.name('enus')} ({self.region_tag.upper()})"

    def get_name(self, locale: Optional[str] = None) -> str:
        """Retrieves the localized name of the :obj:`Region` instance

        Args:
            locale (str, optional): the localization to retrieve. Defaults to locale of instance

        Returns:
            (str) Localized name of the :obj:`Region`
        """
        return self._localized_data('_names', locale)
